import numpy as np
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
import colorlover as cl
from plotly.offline import iplot, init_notebook_mode
import plotly.graph_objs as go
import picturedrocks as pr

_import_errors = None

try:
    import ipywidgets as ipyw
    from IPython.display import display
    from tqdm._tqdm_notebook import tqdm_notebook as tqdm

    tqdm.monitor_interval = False
except ImportError as e:
    _import_errors = e.name


class InteractiveMarkerSelection:
    def __init__(self, adata, infoset, obj, disp_genes=10, connected=True):
        """Run an interactive marker selection GUI inside a jupyter notebook

        Args
        ----
        adata: anndata.AnnData
            The data to run marker selection on. If you want to restrict to a small
            number of genes, slice your anndata object.
        infoset: picturedrocks.markers.InformationSet
            An InformationSet corresponding to `adata`
        obj: function
            An objective function (see `cife_obj` for an example)
        disp_genes: int
            Number of genes to display as options (by default, number of genes
            plotted on the tSNE plot is `3 * disp_genes`, but can be changed by
            setting the `plot_genes` property after initializing.
        connected: bool
            Parameter to pass to `plotly.offline.init_notebook_mode`. If your
            browser does not have internet access, you should set this to False.

        Warning
        -------
        This class requires modules not explicitly listed as dependencies of
        picturedrocks. Specifically, please ensure that you have `ipywidgets` and
        `tqdm` installed and that you use this class only inside a jupyter notebook.
        """

        if _import_errors:
            raise ImportError(f"Unable to import {_import_errors}")
        self.adata = adata
        self.infoset = infoset
        self.obj = obj
        self.S = []
        self.pool = np.arange(adata.n_vars)
        self.scores = np.zeros(len(self.pool))

        init_notebook_mode(connected=connected)

        if not "gene_tsne" in self.adata.varm_keys():
            print("Running tSNE on genes...")
            p = PCA(n_components=30)
            t = TSNE()
            self.adata.varm["gene_pca"] = p.fit_transform(self.adata.X.T)
            self.adata.varm["gene_tsne"] = t.fit_transform(self.adata.varm["gene_pca"])

        if not "X_tsne" in self.adata.obsm_keys():
            print("Running tSNE on cells...")
            p = PCA(n_components=30)
            t = TSNE()
            self.adata.obsm["X_pca"] = p.fit_transform(self.adata.X)
            self.adata.obsm["X_tsne"] = t.fit_transform(self.adata.obsm["X_pca"])

        self.disp_genes = disp_genes
        self.plot_genes = disp_genes * 3

        self.out_next = ipyw.VBox()
        self.out_cur = ipyw.VBox()
        self.out_plot = ipyw.Output()
        self.out = ipyw.Output()

        with self.out:
            display(
                ipyw.VBox([ipyw.HBox([self.out_next, self.out_cur]), self.out_plot])
            )

    def compute_redraw(self):
        """Recompute scores and redraw jupyter widgets"""
        H = self.infoset.H
        self.out_next.children = []
        self.out_cur.children = []

        self.out_plot.clear_output()

        with self.out_plot:
            self.scores = np.array([self.obj(H, i, self.S) for i in tqdm(self.pool)])
        baseline_score = self.obj(H, self.infoset.baseline_index, self.S)
        self.scores -= baseline_score

        top_gene_inds = np.argsort(self.scores)[::-1]

        self.out_next.children = (
            [ipyw.Label("Candidate Next Gene")]
            + [
                self._next_gene_row(gene_ind, self.scores[gene_ind])
                for gene_ind in top_gene_inds[: self.disp_genes]
            ]
            + [self._other_next_gene()]
        )

        self.out_cur.children = [ipyw.Label("Currently selected genes")] + [
            self._cur_gene_row(gene_ind) for gene_ind in self.S
        ]

        ninfscores = self.scores == float("-inf")
        scaled_scores = self.scores - self.scores[~ninfscores].min()
        scaled_scores[scaled_scores < 0] = 0
        scaled_scores = scaled_scores / scaled_scores.max()

        top_genes_plot = top_gene_inds[: self.plot_genes]
        self.out_plot.clear_output()
        with self.out_plot:
            iplot(
                pr.plot.plotgeneheat(
                    self.adata,
                    self.adata.obsm["X_tsne"],
                    top_gene_inds[: self.disp_genes].tolist() + self.S,
                )
            )
            iplot(
                go.Figure(
                    data=[
                        go.Scatter(
                            x=self.adata.varm["gene_tsne"][top_genes_plot, 0],
                            y=self.adata.varm["gene_tsne"][top_genes_plot, 1],
                            text=self.adata.var_names.values.astype(str)[
                                top_genes_plot
                            ],
                            mode="markers",
                            hoverinfo="text",
                            marker=dict(
                                size=(scaled_scores[top_genes_plot] * 16).astype(int),
                                color=np.array(cl.scales["9"]["seq"]["Blues"])[
                                    (3 + 5 * scaled_scores[top_genes_plot].astype(int))
                                ],
                                line=go.scatter.marker.Line(color="black", width=1),
                            ),
                        )
                    ],
                    layout=go.Layout(hovermode="closest"),
                )
            )

    def _next_gene_row(self, gene_ind, score):
        but = ipyw.Button(
            icon="plus",
            tooltip="Add {}".format(self.adata.var_names[gene_ind]),
            layout=dict(width="40px"),
        )

        def add_cur_gene(b):
            self.S.append(gene_ind)
            self.compute_redraw()

        but.on_click(add_cur_gene)
        return ipyw.HBox(
            [
                but,
                ipyw.Label(
                    "{} (score: {:0.4f})".format(self.adata.var_names[gene_ind], score),
                    layout=dict(width="250px"),
                ),
            ]
        )

    def _cur_gene_row(self, gene_ind):
        def del_cur_gene(b):
            self.S.remove(gene_ind)
            self.compute_redraw()

        pop_but = ipyw.Button(
            icon="minus-circle",
            tooltip="Remove {}".format(self.adata.var_names[gene_ind]),
            layout=dict(width="40px"),
        )
        pop_but.on_click(del_cur_gene)
        return ipyw.HBox(
            [
                ipyw.Label(self.adata.var_names[gene_ind], layout=dict(width="250px")),
                pop_but,
            ]
        )

    def _other_next_gene(self):
        but = ipyw.Button(
            icon="plus", tooltip="Add", disabled=True, layout=dict(width="40px")
        )
        label = ipyw.Label("(?)")
        gene_ind = -1
        textbox = ipyw.Text(layout=dict(width="150px"), placeholder="Gene Name")

        def name_updated(change):
            nonlocal gene_ind
            try:
                gene_ind = self.adata.var_names.get_loc(textbox.value)
                label.value = "(score: {:0.4f})".format(self.scores[gene_ind])
                but.disabled = False
            except KeyError:
                gene_ind = -1
                label.value = "(?)"
                but.disabled = True

        textbox.observe(name_updated, names="value")

        def add_other_gene(b):
            if gene_ind >= 0:
                self.S.append(gene_ind)
                self.compute_redraw()

        but.on_click(add_other_gene)
        return ipyw.HBox([but, textbox, label])

    def show(self):
        """Display the jupyter widgets"""
        display(self.out)
        self.compute_redraw()


def cife_obj(H, i, S):
    """The CIFE objective function for feature selection
    
    Args
    ----
    H: function
        an entropy function, typically the bound method H on an instance of
        InformationSet. For example, if `infoset` is of type
        `picturedrocks.markers.InformationSet`, then pass `infoset.H`
    i: int
        index of candidate gene
    S: list
        list of features already selected
    Returns
    -------
    float
        the candidate feature's score relative to the selected gene set `S`
    """
    Sset = set(S)
    m = len(S)
    if i in Sset:
        return float("-inf")
    curobj = (1 - m) * (H((i,)) - H((-1, i)))
    for x in S:
        curobj += H((x, i)) - H((-1, x, i))
    return curobj



# Copyright © 2017, 2018 Anna Gilbert, Alexander Vargo, Umang Varma
#
# This file is part of PicturedRocks.
#
# PicturedRocks is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PicturedRocks is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PicturedRocks.  If not, see <http://www.gnu.org/licenses/>.

from anndata import AnnData
import numpy as np
import colorlover as cl
import plotly.graph_objs as go
import collections
from .preprocessing import pca


def _deep_merge_dict(source, destination):
    """Merges dict-like objects in a deep manner

    e.g., if source = {'a': {'b': 5}, 'c': 4} and destination = {'a': {'b':
    1, 'bb': 0},'d': 0} then destination is changed **in place** to {'a':
    {'b': 5, 'bb': 0}, 'c': 4, 'd': 0}
    """

    for key in source.keys():
        if isinstance(source[key], collections.Mapping) and key in destination:
            _deep_merge_dict(source[key], destination[key])
        else:
            destination[key] = source[key]


def _scatter(coords, *args, **kwargs):
    """Run the appropriate scatter function"""
    assert coords.shape[1] in [2, 3], "incorrect dimensions for coords"
    if coords.shape[1] == 2:
        return go.Scatter(x=coords[:, 0], y=coords[:, 1], *args, **kwargs)
    else:
        return go.Scatter3d(
            x=coords[:, 0], y=coords[:, 1], z=coords[:, 2], *args, **kwargs
        )


def genericplot(celldata, coords, **scatterkwargs):
    """Generate a figure for some embedding of data

    This function supports both 2D and 3D plots. This may be used to plot
    data for any embedding (e.g., PCA or t-SNE). For example usage, see code
    for `pcafigure`.

    Args
    ----
    celldata: anndata.AnnData
        data to plot
    coords: numpy.ndarray
        (N, 2) or (N, 3) shaped coordinates of the embedded data 
    **scatterkwargs
        keyword arguments to pass to ``Scatter`` or ``Scatter3D`` in `plotly`
        (dictionaries are merged recursively)
    """

    clusterindices = celldata.uns["clusterindices"]
    colscal = cl.scales["9"]["qual"]["Set1"]

    plotdata = [
        _scatter(
            coords[inds],
            mode="markers",
            marker=dict(
                size=4,
                color=colscal[k % len(colscal)],  # set color to an array/list
                #                                  of desired values
                opacity=1,
            ),
            name=celldata.obs["clust"].cat.categories[k],
            hoverinfo="name",
        )
        for k, inds in clusterindices.items()
    ]

    if scatterkwargs:
        for p in plotdata:
            _deep_merge_dict(scatterkwargs, p)

    layout = go.Layout(margin=dict(l=0, r=0, b=0, t=0), hovermode="closest")
    return go.Figure(data=plotdata, layout=layout)


def pcafigure(celldata, **scatterkwargs):
    """Make a 3D PCA figure for an AnnData object

    Args
    ----
    celldata: anndata.AnnData
        data to plot
    **scatterkwargs
        keyword arguments to pass to ``Scatter`` or ``Scatter3D`` in `plotly`
        (dictionaries are merged recursively)
    """
    if "X_pca" not in celldata.obsm_keys() or celldata.uns["num_pcs"] < 3:
        print("Need 3 PCs. Calculating now.")
        pca(celldata, 3)
    return genericplot(celldata, celldata.obsm["X_pca"][:, -3:], **scatterkwargs)


def genericwrongplot(celldata, coords, yhat, labels=None, **scatterkwargs):
    """Plot figure with incorrectly classified points highlighted
    
    This can be used with any 2D or 3D embedding (e.g., PCA or t-SNE). For
    example code, see `pcawrongplot`.

    Args
    ----
    celldata: anndata.AnnData
        data to plot
    coords: numpy.ndarray
        (N, 2) or (N, 3) shaped array with coordinates to plot
    yhat: numpy.ndarray
        (N, 1) shaped array of predicted `y` values
    labels: list, optional
        list of axis titles
    **scatterkwargs
        keyword arguments to pass to ``Scatter`` or ``Scatter3D`` in `plotly`
        (dictionaries are merged recursively)
    """

    y = celldata.obs["y"]
    if yhat.shape == (y.shape[0], 1):
        yhat = yhat[:, 0]
    assert yhat.shape == y.shape, "yhat must have shape (N, 1)"

    colscal = cl.scales["9"]["qual"]["Set1"]

    # indices (by cluster) where wrong
    wronginds = {}
    for k in range(celldata.uns["num_clusts"]):
        wronginds[k] = np.nonzero((y == k) & np.not_equal(yhat, y))[0]

    # indices (by cluster) where correct
    clustinds = {}
    for k in range(celldata.uns["num_clusts"]):
        clustinds[k] = np.nonzero((y == k) & np.equal(yhat, y))[0]

    # Get the points that are wrong
    plotdata = [
        _scatter(
            coords[inds],
            mode="markers",
            marker=dict(size=4, color=colscal[k % len(colscal)], opacity=1),
            name="Cluster {}".format(k),
            hoverinfo="name+text",
            text=["Predict {}".format(str(a)) for a in yhat[inds]],
        )
        for k, inds in wronginds.items()
    ] + [
        _scatter(
            coords[inds],
            mode="markers",
            marker=dict(size=4, color=colscal[k % len(colscal)], opacity=0.2),
            name="Cluster {}".format(k),
            hoverinfo="name",
            showlegend=False,
        )
        for k, inds in clustinds.items()
    ]

    if scatterkwargs:
        for p in plotdata:
            _deep_merge_dict(scatterkwargs, p)

    layout = go.Layout(margin=dict(l=0, r=0, b=0, t=0), hovermode="closest")
    if labels:
        layout["scene"] = dict(xaxis=dict(title=labels[0]), yaxis=dict(title=labels[1]))
        if len(labels) == 3:
            layout["scene"]["zaxis"] = dict(title=labels[2])

    return go.Figure(data=plotdata, layout=layout)


def pcawrongplot(celldata, yhat, **scatterkwargs):
    """Generate a 3D PCA figure with incorrectly classified points highlighted

    Args
    ----
    celldata: anndata.AnnData
        data to plot
    yhat: numpy.ndarray
        (N, 1) shaped array of predicted `y` values
    **scatterkwargs
        keyword arguments to pass to ``Scatter`` or ``Scatter3D`` in `plotly`
        (dictionaries are merged recursively)
    """

    if "X_pca" not in celldata.obsm_keys() or celldata.uns["num_pcs"] < 3:
        print("Need 3 PCs. Calculating now.")
        pca(celldata, 3)
    return genericwrongplot(
        celldata,
        celldata.obsm["X_pca"][:, -3:],
        yhat,
        labels=["PC1", "PC2", "PC3"],
        **scatterkwargs
    )


def plotgeneheat(celldata, coords, genes, **scatterkwargs):
    """Generate gene heat plot for some embedding of AnnData

    This generates a figure with multiple dropdown options. The first option
    is "Clust" for a plot similar to `genericplot`, and the remaining
    dropdown options correspond to genes specified in `genes`. When
    `celldata.genes` is defined, these drop downs are labeled with the gene
    names.

    Args
    ----
    celldata: anndata.AnnData
        data to plot
    coords: numpy.ndarray
        (N, 2) or (N, 3) shaped coordinates of the embedded data (e.g., PCA
        or tSNE)
    genes: list
        list of gene indices or gene names
    """

    numclusts = celldata.uns["num_clusts"]
    clusterindices = celldata.uns["clusterindices"]
    clustscal = cl.scales["9"]["qual"]["Set1"]
    genescal = np.array(cl.scales["8"]["seq"]["Blues"])

    plotbyclust = [
        _scatter(
            coords[inds],
            mode="markers",
            marker=dict(
                size=4,
                color=clustscal[k % len(clustscal)],  # set color to an array/list
                #                                  of desired values
            ),
            name=celldata.obs["clust"].cat.categories[k],
            hoverinfo="name",
        )
        for k, inds in clusterindices.items()
    ]

    genes = [g if isinstance(g, (int, np.integer)) else celldata.var.index.get_loc(g) for g in genes]
    genenames = celldata.var.index[genes].tolist()

    geneexpr = celldata.X[:, genes]
    exprnorm = np.round(geneexpr * 7 / geneexpr.max(axis=0)).astype(
        int
    )  # this sets the color of each point
    numgenes = geneexpr.shape[1]

    plotbygene = [
        _scatter(
            coords,
            mode="markers",
            marker=dict(size=4, color=genescal[exprnorm[:, i]]),
            name=genenames[i],
            text=geneexpr[:, i].astype(str),
            hoverinfo="name+text",
            visible=False,
        )
        for i, genename in enumerate(genenames)
    ]
    buttons = [
        dict(
            label="Clust",
            method="update",
            args=[{"visible": [True] * numclusts + [False] * numgenes}, {}],
        )
    ]
    for i, genename in enumerate(genenames):
        v = [False] * (numclusts + numgenes)
        v[numclusts + i] = True
        buttons.append(dict(label=genename, method="update", args=[{"visible": v}, {}]))
    if scatterkwargs:
        for p in plotbyclust + plotbygene:
            _deep_merge_dict(scatterkwargs, p)

    updatemenus = [
        dict(
            buttons=buttons,
            direction="down",
            x=0.1,
            y=1.1,
            xanchor="left",
            yanchor="top",
            pad={"r": 10, "t": 10},
        )
    ]
    layout = go.Layout(
        margin=dict(l=0, r=0, b=0, t=0),
        hovermode="closest",
        updatemenus=updatemenus,
        annotations=[
            dict(
                text="View:",
                x=0.01,
                y=1.065,
                xref="paper",
                yref="paper",
                align="left",
                showarrow=False,
            )
        ],
        showlegend=True,
    )

    return go.Figure(data=plotbyclust + plotbygene, layout=layout)



# Copyright © 2017, 2018 Anna Gilbert, Alexander Vargo, Umang Varma
#
# This file is part of PicturedRocks.
#
# PicturedRocks is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PicturedRocks is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PicturedRocks.  If not, see <http://www.gnu.org/licenses/>.

import pandas as pd

from anndata import AnnData
from warnings import warn
import numpy as np


def read_clusts(adata, filename, sep=",", copy=False):
    """Read cluster labels from a csv

    Args
    ----
    adata: anndata.AnnData
        the `AnnData` object to read labels into
    filename: str
        filename of the csv file with labels
    sep: str, optional
        csv delimiter 
    copy: bool
        determines whether a copy of `AnnData` object is returned

    Returns
    -------
    anndata.AnnData
        object with cluster labels
    
    Notes
    -----
     * Cluster ids will automatically be changed so they are 0-indexed
     * csv can either be two columns (in which case the first column is treated
       as observation label and merging handled by pandas) or one column (only
       cluster labels, ordered as in ``adata``)
    """
    adata = adata.copy() if copy else adata
    clustdf = pd.read_csv(filename, sep=sep)
    if clustdf.shape[1] == 2:
        clustdf = clustdf.set_index(clustdf.columns[0])
    assert clustdf.shape[1] == 1, "Cluster column ambigious"
    clusters = clustdf.iloc[:, 0]
    if clusters.dtype.kind == "i":
        if clusters.min() > 0:
            warn("Changing cluster ids to begin at 0.")
            clusters -= clusters.min()
        clustuniq = np.sort(clusters.unique())
        assert np.array_equal(
            clustuniq, np.arange(clustuniq.size)
        ), "Cluster ids need to be 0, 1, ..., K-1"
        adata.obs["y"] = clusters
        adata.obs["clust"] = ("Cluster " + clusters.astype("str")).astype("category")
    else:
        adata.obs["clust"] = clusters.astype("category")
        adata.obs["y"] = adata.obs["clust"].cat.codes
    if adata.obs["y"].isnull().any() or adata.obs["clust"].isnull().any():
        warn("Some or all cells not assigned to cluster.")
    return adata


def process_clusts(adata, copy=False):
    """Annotate with information about clusters

    Precomputes cluster indices, number of clusters, etc.

    Args
    ----
    adata: anndata.AnnData
    copy: bool
        determines whether a copy of `AnnData` object is returned

    Return
    -------
    anndata.AnnData
        object with annotation

    Notes
    ------
    The information computed here is lost when saving as a `.loom` file. If a
    `.loom` file has cluster information, you should run this function
    immediately after :func:`sc.read_loom <scanpy.api.read_loom>`.
    """
    adata = adata.copy() if copy else adata
    adata.obs["clust"] = adata.obs["clust"].astype("category")
    adata.uns["num_clusts"] = adata.obs["clust"].cat.categories.size
    clusterindices = {}
    for k in range(adata.uns["num_clusts"]):
        clusterindices[k] = (adata.obs["y"] == k).nonzero()[0]
    adata.uns["clusterindices"] = clusterindices
    return adata



# Copyright © 2017, 2018 Anna Gilbert, Alexander Vargo, Umang Varma
#
# This file is part of PicturedRocks.
#
# PicturedRocks is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PicturedRocks is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PicturedRocks.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import scipy.sparse
import datetime
from logging import info


def makeinfoset(adata):
    """Discretize data"""
    # we currently don't support scipy sparse matrices
    X = adata.X.toarray() if scipy.sparse.issparse(adata.X) else adata.X
    return InformationSet(np.log2(X + 1).round().astype(int), adata.obs["y"])


def mutualinfo(infoset, n, pool=None, obj="mrmr"):
    """Compute markers using mutual information

    Args
    ----
    infoset: picturedrocks.markers.InformationSet
        discrete gene expression matrix
    n: int
        number of markers to select
    pool: list, optional
        pool of genes to restrict marker selection search 
    obj: {"mrmr", "mim", "cife", "jmi", "deg3"}
        objective function to use
    
    Returns
    -------
    list
        ``n`` element list of markers
    """

    assert infoset.hasy, "infoset must include target (cluster) labels"

    if pool is None:
        pool = range(infoset.X.shape[1] - 1)

    start = datetime.datetime.now()
    S = []
    Sset = set()
    H = infoset.H
    for m in range(n):
        maxobj = float("-inf")
        argmaxobj = -1
        if obj == "mim" or m == 0:
            # all objectives do the same thing when m = 0, but some
            # objectives simplify greatly m is not 0, so handle that here
            for i in pool:
                if i in Sset:
                    continue
                curobj = H((i,)) - H((-1, i))
                maxobj, argmaxobj = max((maxobj, argmaxobj), (curobj, i))
        elif obj == "mrmr":
            for i in pool:
                if i in Sset:
                    continue
                curobj = -m * H((-1, i))  # infoset.Iwrtx([-1], i, 2)
                for x in S:
                    curobj += H((x, i))  # (1.0/m) * infoset.Iwrtx([x], i, 3)
                maxobj, argmaxobj = max((maxobj, argmaxobj), (curobj, i))
        elif obj == "jmi":
            for i in pool:
                if i in Sset:
                    continue
                curobj = 0
                for x in S:
                    curobj += H((x, i)) - H((-1, x, i))
                maxobj, argmaxobj = max((maxobj, argmaxobj), (curobj, i))
        elif obj == "cife":
            for i in pool:
                if i in Sset:
                    continue
                curobj = (1 - m) * (H((i,)) - H((-1, i)))
                for x in S:
                    curobj += H((x, i)) - H((-1, x, i))
                maxobj, argmaxobj = max((maxobj, argmaxobj), (curobj, i))
        elif obj == "deg3":
            mchoose2 = (m * (m - 1)) // 2
            for i in pool:
                if i in Sset:
                    continue
                curobj = (1 - m + mchoose2) * (H((i,)) - H((-1, i)))
                for x in S:
                    curobj += (2 - m) * (H((x, i)) - H((-1, x, i)))
                for x1, x2 in combinations(S, 2):
                    curobj += H((x1, x2, i)) - H((-1, x1, x2, i))
                maxobj, argmaxobj = max((maxobj, argmaxobj), (curobj, i))
        else:
            raise ValueError("Unknown objective function.")
        # common for all objectives
        S.append(argmaxobj)
        Sset.add(argmaxobj)
        info("Features: {}".format(repr(S)))
    end = datetime.datetime.now()
    timedelta = end - start
    info(
        "It took {:.2f} minutes to find {} features via {}.".format(
            timedelta.total_seconds() / 60, n, obj
        )
    )
    return S


class InformationSet:
    """Stores discrete gene expression matrix

    Args
    ----
    X: numpy.ndarray
        a (num_obs, num_vars) shape array with ``dtype`` :class:`int`
    y: numpy.ndarray, optional
        a (num_obs,) shape array
    """

    def __init__(self, X, y=None):
        self.hasy = not y is None
        # insert a column of zeros; this is our baseline (i.e., it should have a
        # score of zero always)
        X = np.c_[X, np.zeros((X.shape[0], 1))]
        self.baseline_index = -1
        if self.hasy:
            self.X = np.c_[X, y]
            self.baseline_index -= 1
        else:
            self.X = X
        self.N = X.shape[0]
        self._H = {}
        self._shift = int(np.log2(X.max()) + 1)

    def cols(self, cols):
        """Ensemble of columns
        
        This function discretizes the ensemble of columns into
        one column. 
        """
        ncols = len(cols)
        return self.X[:, cols].dot(2 ** (self._shift * np.arange(ncols - 1, -1, -1)))

    def H(self, cols):
        """Entropy of an ensemble of columns
        
        Returns `H(cols)`, where `cols` is treated as an ensemble.
        """
        if cols not in self._H:
            vals, counts = np.unique(self.cols(cols), return_counts=True)
            p = counts / self.N
            self._H[cols] = np.sum(-p * np.log2(p))
        return self._H[cols]

    def Iapprox(self, cols, deg):
        """Approximate multivariate MI
        
        If `cols = [x_1, ..., x_m]`, this function returns the
        degree `deg` approximation of `I(x_1; ...; x_m)`.
        """
        ret = 0
        for k in range(deg):
            for U in combinations(cols, k + 1):
                ret += ((-1) ** k) * self.H(U)
        return ret

    def Iwrtx(self, S, x, deg):
        """Approximate multivariate MI with resp to x
        
        If `S = [x_1, x_2, ..., x_m]`, this fuction returns the
        sum of the terms in the inclusion/exclusion expansion for
        `I(x_1; x_2; ...; x_m; x)` of degree `deg` that involve `x`.
        
        This is a faster substitute for Iapprox for calculations where
        only the terms involve `x` are required (e.g., to compute
        `argmax_x (x_1;...;x_n; x)`
        """
        ret = 0
        for k in range(deg):
            for U in combinations(S, k):
                ret += ((-1) ** k) * self.H(U + (x,))
        return ret



# Copyright © 2017 Anna Gilbert, Alexander Vargo, Umang Varma
#
# This file is part of PicturedRocks.
#
# PicturedRocks is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PicturedRocks is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PicturedRocks.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import plotly.graph_objs as go
import scanpy.api as sc
import scipy.spatial.distance
from anndata import AnnData

from .read import process_clusts


def kfoldindices(n, k, random=False):
    """Generate indices for k-fold cross validation

    Args
    ----
    n: int
        number of observations
    k: int
        number of folds
    random: bool
        determines whether to randomize the order

    Yields
    ------
    numpy.ndarray
        array of indices in each fold
    """
    basearray = np.arange(n)
    if random:
        np.random.shuffle(basearray)
    lengthfloor = n // k
    extra = n % k
    cur = 0
    while cur < n:
        thislength = lengthfloor + 1 if extra > 0 else lengthfloor
        yield basearray[cur : cur + thislength]
        cur += thislength
        extra -= 1  # this should be extra = max(extra - 1, 0),
        #            but it doesn't matter


class PerformanceReport:
    """Report actual vs predicted statistics
    
    Args
    ----
    y: numpy.ndarray
        actual cluster labels, (N, 1)-shaped numpy array
    yhat: numpy.ndarray
        predicted cluster labels, (N, 1)-shaped numpy array
    """

    def __init__(self, y, yhat):
        self.y = y
        self.yhat = yhat
        self.N = y.shape[0]

        self.K = y.max() + 1
        assert np.equal(
            np.unique(self.y), range(self.K)
        ).all(), "Cluster labels should be 0, 1, 2, ..., K -1"

        self.clusterindices = {}
        self._genclusterindices()

    def _genclusterindices(self):
        """Compute and store indices for cells in each cluster."""
        for k in range(self.K):
            self.clusterindices[k] = np.nonzero(self.y == k)[0]
        self.nk = np.array([len(self.clusterindices[k]) for k in range(self.K)])
        # nk[k] is the number of entries in cluster k

    def wrong(self):
        """Returns the number of cells misclassified."""
        return np.sum((self.y.flatten() != self.yhat) * 1.0)

    def printscore(self):
        """Print a message with the score"""
        wrong = self.wrong()
        print(
            "{} out of {} incorrect: {:.2f}%".format(
                wrong, self.N, 100 * wrong / self.N
            )
        )

    def getconfusionmatrix(self):
        """Get the confusion matrix for the latest run
        
        Returns
        -------
        numpy.ndarray
            array of shape (K, K), with the [i, j] entry being the fraction
            of cells in cluster i that were predicted to be in cluster j
        """
        K = self.K
        freq_table = np.zeros([K, K])
        for i in range(K):
            clust, clust_count = np.unique(
                self.yhat[self.clusterindices[i]], return_counts=True
            )
            for j, k in enumerate(clust):
                freq_table[i, k] = clust_count[j] / self.nk[i]
        return freq_table

    def confusionmatrixfigure(self):
        """Compute and make a confusion matrix figure
        
        Returns
        -------
        `plotly figure`
            confusion matrix 
        """
        freq_table = self.getconfusionmatrix()
        shape = freq_table.shape
        trace = go.Heatmap(
            z=freq_table,
            x=np.arange(shape[1]),
            y=np.arange(shape[0]),
            colorscale="Greys",
            reversescale=True,
        )
        layout = go.Layout(
            title="Confusion Matrix",
            xaxis=dict(title="Predicted Cluster"),
            yaxis=dict(title="Actual Cluster", scaleanchor="x"),
            width=450,
            height=450,
            margin=go.layout.Margin(l=70, r=70, t=70, b=70, pad=0, autoexpand=False),
            annotations=[
                dict(
                    text="Rows sum to 1",
                    x=0.5,
                    y=1,
                    xref="paper",
                    yref="paper",
                    xanchor="center",
                    yanchor="bottom",
                    showarrow=False,
                )
            ],
        )
        return go.Figure(data=[trace], layout=layout)

    def show(self):
        """Print a full report
        
        This uses `iplot`, so we assume this will only be run in a Jupyter
        notebook and that `init_notebook_mode` has already been run.
        """
        from plotly.offline import iplot
        self.printscore()
        iplot(self.confusionmatrixfigure())


class FoldTester:
    """Performs K-fold Cross Validation for Marker Selection

    :class:`FoldTester` can be used to evaluate various marker selection
    algorithms. It can split the data in `K` folds, run marker selection
    algorithms on these folds, and classify data based on testing and
    training data.

    Args
    ----
    adata: anndata.AnnData
        data to slice into folds
    """

    def __init__(self, adata):
        self.adata = adata

        self.folds = None
        self.yhat = None
        self.markers = None

    def makefolds(self, k=5, random=False):
        """Makes folds

        Args
        ----
        k: int
            the value of K
        random: bool
            If true, `makefolds` will make folds randomly. Otherwise, the
            folds are made in order (i.e., the first ``ceil(N / k)`` cells in
            the first fold, etc.)
        """
        self.folds = list(kfoldindices(self.adata.n_obs, k, random))

    def savefolds(self, file):
        """Save folds to a file

        Args
        ----
        file: str
            filename to save (typically with a ``.npz`` extension)
        """
        d = {"k": len(self.folds), "y": self.adata.obs["y"].values}
        for i, f in enumerate(self.folds):
            d["fold{}".format(i)] = f
        return np.savez(file, **d)

    def loadfolds(self, file):
        """Load folds from a file

        The file can be one saved either by :meth:`FoldTester.savefolds` or
        :meth:`FoldTester.savefoldsandmarkers`. In the latter case, it will
        not load any markers.

        See Also
        --------
        :meth:`FoldTester.loadfoldsandmarkers`.
        """
        d = np.load(file)
        k = d["k"]
        self.folds = [d["fold{}".format(i)] for i in range(k)]
        assert np.array_equal(
            self.adata.obs["y"], d["y"].ravel()
        ), "y vector does not match."
        assert self.validatefolds(), "folds are not partition of indices"

    def validatefolds(self):
        """Ensure that all observations are in exactly one fold
        
        Returns
        -------
        bool
        """
        counts = np.zeros(self.adata.n_obs)
        for f in self.folds:
            counts[f] += 1
        return np.alltrue(counts == 1)

    def selectmarkers(self, select_function):
        """Perform a marker selection algorithm on each fold

        Args
        ----
        select_function: function
            a function that takes in an :class:`AnnData <anndata.AnnData>`
            object and outputs a list of gene markers, given by their index 
        """
        k = len(self.folds)
        self.markers = []
        for f in self.folds:
            mask = np.zeros(self.adata.n_obs, dtype=bool)
            mask[f] = True
            traindata = self.adata[~mask, :].copy()
            self.markers.append(select_function(traindata))

    def classify(self, classifier):
        """Classify each cell using training data from other folds

        For each fold, we project the data onto the markers selected for that
        fold, which we treat as test data. We also project the complement of
        the fold and treat that as training data.

        Args
        ----
        classifier
            a classifier that trains with a training data set and predicts
            labels of test data. See `NearestCentroidClassifier` for an
            example.
        """
        self.yhat = np.zeros(self.adata.n_obs, dtype=int) - 1
        for i, f in enumerate(self.folds):
            mask = np.zeros(self.adata.n_obs, dtype=bool)
            mask[f] = True
            traindata = self.adata[~mask, :][:, self.markers[i]].copy()
            c = classifier()
            c.train(traindata)
            self.yhat[f] = c.test(self.adata.X[f, :][:, self.markers[i]].copy())

    def savefoldsandmarkers(self, file):
        """Save folds and markers for each fold

        This saves folds, and for each fold, the markers previously found by
        :meth:`FoldTester.selectmarkers`.

        Args
        ----
        file: str
            filename to save to (typically with a ``.npz`` extension)
        """
        d = {"k": len(self.folds), "y": self.adata.obs["y"]}
        for i, f in enumerate(self.folds):
            d["fold{}".format(i)] = f
        for i, m in enumerate(self.markers):
            d["marker{}".format(i)] = m
        return np.savez(file, **d)

    def loadfoldsandmarkers(self, file):
        """Load folds and markers

        Loads a folds and markers file saved by
        :meth:`FoldTester.savefoldsandmarkers`

        Args
        ----
        file: str
            filename to load from (typically with a ``.npz`` extension)
        
        See Also
        --------
        :meth:`FoldTester.loadfolds`
        """
        d = np.load(file)
        k = d["k"]
        self.folds = [d["fold{}".format(i)] for i in range(k)]
        self.markers = [d["marker{}".format(i)] for i in range(k)]
        assert np.array_equal(
            self.adata.obs["y"], d["y"].ravel()
        ), "y vector does not match."
        assert self.validatefolds(), "folds are not partition of indices"


class NearestCentroidClassifier:
    """Nearest Centroid Classifier for Cross Validation

    Computes the centroid of each cluster label in the training data, then
    predicts the label of each test data point by finding the nearest centroid.
    """

    def __init__(self):
        self.xkibar = None

    def train(self, adata):
        adata = process_clusts(adata)
        adata = sc.pp.normalize_per_cell(adata, 1000, copy=True)
        adata = sc.pp.log1p(adata, copy=True)
        self.xkibar = np.array(
            [
                adata.X[adata.uns["clusterindices"][k]].mean(axis=0)
                for k in range(adata.uns["num_clusts"])
            ]
        )

    def test(self, Xtest):
        testdata = AnnData(Xtest)
        testdata = sc.pp.normalize_per_cell(testdata, 1000, copy=True)
        testdata = sc.pp.log1p(testdata, copy=True)
        dxixk = scipy.spatial.distance.cdist(testdata.X, self.xkibar)
        return dxixk.argmin(axis=1)



# Copyright © 2017, 2018 Anna Gilbert, Alexander Vargo, Umang Varma
#
# This file is part of PicturedRocks.
#
# PicturedRocks is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PicturedRocks is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PicturedRocks.  If not, see <http://www.gnu.org/licenses/>.

from anndata import AnnData
from scipy.sparse.linalg import svds
import numpy as np


def pca(data, dim=3, center=True, copy=False):
    """Runs PCA

    Args
    ----
    data: anndata.AnnData
        input data
    dim: int, optional
        number of PCs to compute
    center: bool, optional
        determines whether to center data before running PCA
    copy: 
        determines whether a copy of `AnnData` object is returned

    Returns
    -------
    anndata.AnnData
        object with ``obsm["X_pca"]``, ``varm["PCs"]``, and ``uns["num_pcs"]`` set
    """

    if isinstance(data, AnnData):
        adata = data.copy() if copy else data
        Xcent, pcs, Xpca = pca(adata.X, dim=dim, center=center)
        adata.obsm["X_pca"] = Xpca
        adata.varm["PCs"] = pcs
        adata.uns["num_pcs"] = dim
        return adata
    Xcent = data.copy()
    if center:
        Xcent -= data.mean(axis=0)[np.newaxis, :]
    pcs = svds(Xcent.T, dim, return_singular_vectors="u")[0][:, ::-1]
    Xpca = Xcent.dot(pcs)
    return (Xcent, pcs, Xpca)



import setuptools
import versioneer

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="picturedrocks",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Anna Gilbert, Alexander Vargo, Umang Varma",
    author_email="uvarma@umich.edu",
    description="Single Cell RNA Sequencing Marker Selection Package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/umangv/picturedrocks",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "anndata",
        "colorlover",
        "h5py",
        "loompy",
        "numpy",
        "pandas",
        "plotly",
        "scanpy",
        "scipy",
    ],
    python_requires="~=3.6",
)



from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

name = "picturedrocks"

from . import read
from . import preprocessing
from . import plot
from . import markers
from . import performance
